@@ -3,6 +3,7 @@
 If it has not been initialized, VTES will evaluate to False.
 VTES must be configured with `VTES.configure()` before being used.
 """
+import collections
 import csv
 import functools
 import difflib
@@ -10,38 +11,105 @@ import io
 import itertools
 import logging
 import pickle
+import re
 import requests
 import tempfile
 import textwrap
+import warnings
 import zipfile
 
 import unidecode
+import yaml
 
 from . import config
 
 logger = logging.getLogger()
 
+COMPLETION_TRIE = collections.defaultdict(lambda: collections.defaultdict(int))
+
+
+def _get_runling_links(links_dict, text):
+    """Yield ruling references and links from a dict of links and a ruling text
+
+    Args:
+        links_dict (dict): {reference: url} of rulings
+        text (str): A ruling with one or more ruling references like [ZZZ 20000101]
+    """
+    references = re.findall(r"\[[a-zA-Z]+\s[0-9-]+\]", text)
+    if not references:
+        warnings.warn(f"no reference in ruling: {text}")
+    for reference in references:
+        reference = reference[1:-1]
+        yield reference, links_dict[reference]
+
 
 class _VTES(dict):
     """VTES cards database that matches incomplete or misspelled card names.
 
-    Keys are lower case card names.
+    Keys are lower case card names and card Ids.
     Contains both crypt and library cards in a single dict.
     """
 
-    def load_from_vekn(self, save=True):
-        """Load the card database from vekn.net
+    def _yaml_get_card(self, text):
+        """Get a card from a YAML card reference (id|Name)
         """
+        card_id, card_name = text.split("|")
+        card_id = int(card_id)
+        if self.get_name(self[card_id]) != card_name:
+            warnings.warn(
+                f"rulings: {card_id} listed as {card_name} "
+                f"instead of {self.get_name(self[card_id])}"
+            )
+        return self[card_id]
+
+    def load_from_vekn(self, save=True, safe_variants=True):
+        """Load the card database from vekn.net, with rulings
+
+        Args:
+            save (bool): If True, card list is pickled to be retrieved faster later on.
+            safe_variants (bool): If True, use more card name variants
+                                  (unsafe for TWDA parsing)
+        """
+        global COMPLETION_TRIE
         self.clear()
+        COMPLETION_TRIE.clear()
         r = requests.request("GET", config.VEKN_VTES_URL)
+        r.raise_for_status()
         with tempfile.NamedTemporaryFile("wb", suffix=".zip") as f:
             f.write(r.content)
             f.flush()
             z = zipfile.ZipFile(f.name)
             with z.open(config.VEKN_VTES_LIBRARY_FILENAME) as c:
-                self.load_csv(io.TextIOWrapper(c, encoding="utf_8_sig"), save)
+                self.load_csv(
+                    io.TextIOWrapper(c, encoding="utf_8_sig"), save, safe_variants
+                )
             with z.open(config.VEKN_VTES_CRYPT_FILENAME) as c:
-                self.load_csv(io.TextIOWrapper(c, encoding="utf_8_sig"), save)
+                self.load_csv(
+                    io.TextIOWrapper(c, encoding="utf_8_sig"), save, safe_variants
+                )
+        # load rulings from local yaml files
+        links = yaml.safe_load(open(config.RULINGS_LINK_FILE, "r"))
+        cards_rulings = yaml.safe_load(open(config.CARDS_RULINGS_FILE, "r"))
+        general_rulings = yaml.safe_load(open(config.GENERAL_RULINGS_FILE, "r"))
+        for card, rulings in cards_rulings.items():
+            card = self._yaml_get_card(card)
+            card.setdefault("Rulings", []).extend(rulings)
+            card.setdefault("Rulings Links", [])
+            for i, ruling in enumerate(rulings):
+                for reference, link in _get_runling_links(links, ruling):
+                    card.setdefault("Rulings Links", []).append(
+                        {"Reference": reference, "URL": link}
+                    )
+        for ruling in general_rulings:
+            for card in ruling["cards"]:
+                card = self._yaml_get_card(card)
+                card.setdefault("Rulings", []).append(ruling["ruling"])
+                card.setdefault("Rulings Links", [])
+                for reference, link in _get_runling_links(links, ruling["ruling"]):
+                    card.setdefault("Rulings Links", []).append(
+                        {"Reference": reference, "URL": link}
+                    )
+        pickle.dump(self, open(config.VTES_FILE, "wb"))
 
     def __getitem__(self, key):
         """Get a card, try to find a good matching.
@@ -52,7 +120,8 @@ class _VTES(dict):
         Args:
             key (str): the card to search for
         """
-        key = key.lower()
+        if isinstance(key, str):
+            key = key.lower()
         try:
             return super().__getitem__(key)
         except KeyError:
@@ -67,16 +136,17 @@ class _VTES(dict):
     def __contains__(self, key):
         """Also use config.REMAP and difflib on "in" tests
         """
-        key = key.lower()
+        if isinstance(key, str):
+            key = key.lower()
         return (
             super().__contains__(key) or key in config.REMAP or self._fuzzy_match(key)
         )
 
     def __setitem__(self, key, value):
-        return super().__setitem__(key.lower(), value)
+        return super().__setitem__(key.lower() if isinstance(key, str) else key, value)
 
     def __delitem__(self, key, value):
-        return super().__setitem__(key.lower(), value)
+        return super().__setitem__(key.lower() if isinstance(key, str) else key, value)
 
     def configure(self):
         """Add alternative names for some cards
@@ -86,9 +156,40 @@ class _VTES(dict):
         for alternative, card in config.AKA.items():
             self[alternative] = self[card]
 
+        # build a completion trie for card names (use in web API and discord bot)
+        COMPLETION_TRIE.clear()
+        for name, card in self.items():
+            if isinstance(name, int):
+                continue
+            for e, part in enumerate(name.lower().split()):
+                for i in range(1, len(part) + 1):
+                    COMPLETION_TRIE[part[:i]][self.get_name(card)] += (
+                        # double score for matching name start
+                        i
+                        * (2 if e == 0 else 1)
+                    )
+
+    def complete(self, partial_name):
+        ret = None
+        for part in partial_name.split():
+            part_match = {}
+            for match, score in COMPLETION_TRIE.get(part.lower(), dict()).items():
+                part_match[match] = max(part_match.get(match, 0), score)
+            if ret:
+                ret = collections.Counter(
+                    {k: v + part_match[k] for k, v in ret.items() if k in part_match}
+                )
+            else:
+                ret = collections.Counter(part_match)
+        return [x[0] for x in sorted(ret.items(), key=lambda x: (-x[1], x[0]),)]
+
     def __hash__(self):
         return id(self)
 
+    @functools.lru_cache()
+    def all_card_names_variants(self):
+        return [k for k in self.keys() if isinstance(k, str)]
+
     @functools.lru_cache()
     def _fuzzy_match(self, name):
         """Use difflib to match incomplete or misspelled names
@@ -101,8 +202,12 @@ class _VTES(dict):
         Returns:
             The matched card name, or None
         """
+        if not isinstance(name, str):
+            return
         # 0.8 is an empirical value, seems to work nicely
-        result = difflib.get_close_matches(name, self.keys(), n=1, cutoff=0.8)
+        result = difflib.get_close_matches(
+            name, self.all_card_names_variants(), n=1, cutoff=0.8
+        )
         if result:
             match = result[0]
             logger.info("misspelled [{}] matched [{}]".format(name, match))
@@ -144,14 +249,12 @@ class _VTES(dict):
             All possible values
         """
 
-        def get_traits(trait_value):
-            if "/" in trait_value:
-                for value in trait_value.split("/"):
-                    yield value.strip()
-            elif "&" in trait_value:
-                yield "Combo"
-            else:
-                yield trait_value
+        def get_traits(traits):
+            for trait_value in traits:
+                if "&" in trait_value:
+                    yield "Combo"
+                else:
+                    yield trait_value
 
         return set(
             itertools.chain.from_iterable(
@@ -181,14 +284,15 @@ class _VTES(dict):
     def get_name(card):
         """Returns main name for a card
         """
-        advanced_suffix = " (ADV)" if card.get("Adv") else ""
         name = card["Name"]
         if name[-5:] == ", The":
             name = "The " + name[:-5]
-        return name + advanced_suffix
+        if name[-6:] != " (ADV)" and card.get("Adv"):
+            name = name + " (ADV)"
+        return name
 
     @staticmethod
-    def get_name_variants(card):
+    def get_name_variants(card, safe=True):
         """Yields all name variants for a card
 
         Add the " (ADV)" suffix for advanced vampires
@@ -203,6 +307,7 @@ class _VTES(dict):
 
         Args:
             card (dict): the card
+            safe (bool): if True, yield more variants (unsafe for TWDA parsing)
 
         Yields:
             str: all name variants
@@ -235,7 +340,7 @@ class _VTES(dict):
                 # We do not shorten too much or anything will fuzzy match
                 # eg. "Line" (without ", The") could mismatch "Redline" or "Zip Line".
                 # Still, "Gunther" should match "Gunther, Beast Lord'
-                if len(alternative) > 6:
+                if not safe or len(alternative) > 6:
                     yield from name_variants(alternative)
 
         yield from name_variants(card["Name"])
@@ -243,14 +348,38 @@ class _VTES(dict):
         for alias in card.get("Aka", "").split(";"):
             yield from name_variants(alias.strip())
 
-    def load_csv(self, stream, save=True):
+    def load_csv(self, stream, save=True, safe_variants=True):
         """Load a crypt or library CSV (VEKN format)
 
         The VTES card list is pickled for future use.
         """
         for card in csv.DictReader(stream):
-            for name in self.get_name_variants(card):
+            card["Set"] = [s.strip() for s in card["Set"].split(",") if s.strip()]
+            card["Clan"] = [c.strip() for c in card["Clan"].split("/") if c.strip()]
+            card["Type"] = [t.strip() for t in card["Type"].split("/") if t.strip()]
+            card["Card Text"] = card["Card Text"].replace("{", "")
+            card["Card Text"] = card["Card Text"].replace("}", "")
+            card["Card Text"] = card["Card Text"].replace("(D)", "Ⓓ ")
+            if "Disciplines" in card:
+                card["Disciplines"] = [
+                    d.strip() for d in card["Disciplines"].split(" ") if d.strip()
+                ]
+            if "Requirement" in card:
+                card["Requirement"] = [
+                    r.strip() for r in card["Requirement"].split(",") if r.strip()
+                ]
+            if "Discipline" in card:
+                card["Discipline"] = [
+                    d.strip() for d in card["Discipline"].split("/") if d.strip()
+                ]
+            if "Burn Option" in card:
+                card["Burn Option"] = bool(card["Burn Option"])
+            if "Adv" in card:
+                card["Adv"] = bool(card["Adv"])
+            self[int(card["Id"])] = card
+            for name in self.get_name_variants(card, safe_variants):
                 self[name] = card
+            card["Name"] = self.get_name(card)
 
         # pickle this
         if save:
@@ -266,7 +395,7 @@ class _VTES(dict):
         Returns:
             bool: true if the card is a crypt card
         """
-        return VTES[card]["Type"] in ("Vampire", "Imbued")
+        return set(VTES[card]["Type"]) & {"Vampire", "Imbued"}
 
     def is_library(self, card):
         """A function to check if a card is a library card
@@ -292,10 +421,10 @@ class _VTES(dict):
             bool: true if the card is of the given discipline
         """
         if discipline in ["Combo", "combo"]:
-            return "&" in VTES[card].get("Discipline", "")
+            return any("&" in d for d in VTES[card].get("Discipline", []))
         return {discipline.strip().lower()} & {
             disc.strip().lower()
-            for d in VTES[card].get("Discipline", "").split("/")
+            for d in VTES[card].get("Discipline", [])
             for disc in d.split("&")
         }
 
@@ -308,7 +437,7 @@ class _VTES(dict):
         Returns:
             bool: true if the card requires no discipline
         """
-        return VTES[card].get("Discipline", "").strip() == ""
+        return not VTES[card].get("Discipline", [])
 
     def is_type(self, card, type_):
         """A function to check if a card is of the given type
@@ -322,9 +451,7 @@ class _VTES(dict):
         Returns:
             bool: true if the card has the given type
         """
-        return {type_.strip().lower()} & {
-            t.strip().lower() for t in VTES[card]["Type"].split("/")
-        }
+        return {type_.strip().lower()} & {t.strip().lower() for t in VTES[card]["Type"]}
 
     def is_clan(self, card, clan):
         """A function to check if a card is of the given clan (or requires said clan)
@@ -338,9 +465,7 @@ class _VTES(dict):
         Returns:
             bool: true if the card is of the given clan.
         """
-        return {clan.strip().lower()} & {
-            t.strip().lower() for t in VTES[card]["Clan"].split("/")
-        }
+        return {clan.strip().lower()} & {t.strip().lower() for t in VTES[card]["Clan"]}
 
     def deck_to_txt(self, deck):
         """A consistent deck display matching our parsing rules of TWDA.html
@@ -355,7 +480,7 @@ class _VTES(dict):
         """
 
         def _type(card):
-            return config.TYPE_ORDER.index(self[card[0]]["Type"])
+            return config.TYPE_ORDER.index("/".join(self[card[0]]["Type"]))
 
         lines = []
         if deck.event:
@@ -392,8 +517,8 @@ class _VTES(dict):
                     count,
                     card + (" (ADV)" if self[card]["Adv"] else ""),
                     self[card]["Capacity"],
-                    self[card]["Disciplines"],
-                    self[card]["Clan"],
+                    " ".join(self[card]["Disciplines"]),
+                    self[card]["Clan"][0],
                     self[card]["Group"],
                 )
             )
@@ -417,11 +542,23 @@ class _VTES(dict):
                     lines.append(f"{count:<2} {card}")
         return "\n".join(lines)
 
-    def deck_to_dict(self, deck):
+    def deck_to_dict(self, deck, twda_id):
+        """A consistent deck serialization for the API
+
+        Cards are listed in the order given by the config.TYPE_ORDER list.
+
+        Args:
+            deck (deck.Deck): A deck
+            twda_id (str): The deck TWDA ID
+
+        Returns:
+            dict: The serialized deck
+        """
         ret = {
+            "twda_id": twda_id,
             "event": deck.event,
             "place": deck.place,
-            "date": deck.date.format("YYYY, MMMM Do"),
+            "date": deck.date.date().isoformat(),
             "tournament_format": deck.tournament_format,
             "players_count": deck.players_count,
             "player": deck.player,
@@ -429,15 +566,23 @@ class _VTES(dict):
             "name": deck.name,
             "author": deck.author,
             "comments": deck.comments,
-            "crypt": [
-                deck.cards_count(self.is_crypt),
-                [[count, card] for card, count in deck.cards(self.is_crypt)],
-            ],
-            "library": [deck.cards_count(self.is_library), {}],
+            "crypt": {
+                "count": deck.cards_count(self.is_crypt),
+                "cards": [
+                    {
+                        "id": self[card]["Id"],
+                        "count": count,
+                        "name": card,
+                        "comments": deck.cards_comments.get(card),
+                    }
+                    for card, count in deck.cards(self.is_crypt)
+                ],
+            },
+            "library": {"count": deck.cards_count(self.is_library), "cards": []},
         }
 
         def _type(card):
-            return config.TYPE_ORDER.index(self[card[0]]["Type"])
+            return config.TYPE_ORDER.index("/".join(self[card[0]]["Type"]))
 
         # sort by type, count (descending), name
         # note ordering must match the `itertools.groupby` function afterwards.
@@ -447,12 +592,22 @@ class _VTES(dict):
         # form a section for each type with a header displaying the total
         for kind, cards in itertools.groupby(library_cards, key=_type):
             c1, c2 = itertools.tee(cards)
-            ret["library"][1][config.TYPE_ORDER[kind]] = [
-                sum(count for card, count in c1),
-                [],
-            ]
+            ret["library"]["cards"].append(
+                {
+                    "type": config.TYPE_ORDER[kind],
+                    "count": sum(count for card, count in c1),
+                    "cards": [],
+                }
+            )
             for card, count in c2:
-                ret["library"][1][config.TYPE_ORDER[kind]][1].append([count, card])
+                ret["library"]["cards"][-1]["cards"].append(
+                    {
+                        "id": self[card]["Id"],
+                        "count": count,
+                        "name": card,
+                        "comments": deck.cards_comments.get(card),
+                    }
+                )
         return ret
 
 
